import pandas as pd
from sklearn.model_selection import cross_validate
import matplotlib.pyplot as plt

def calculate_lift(ytrue, yproba, cumulative=True):
    cmp = pd.DataFrame(yproba, columns=['pred'])
    cmp.loc[:, 'y'] = ytrue

    last_value = 1
    deciles = np.linspace(0, 1, 11)[::-1][1:]
    r = []
    for k in deciles:
        value = cmp.pred.quantile(k)
        if cumulative:
            selection = cmp.loc[cmp.pred >= value, :]
        else:
            selection = cmp.loc[(cmp.pred >= value) &
                                (cmp.pred <= last_value), :]
            last_value = value
        num_true = (selection.y == 1).sum()
        num_total = selection.shape[0]
        rprate = num_true / num_total
        r.append(dict(cutoff=value,
                      decile=k,
                      responserate=rprate,
                      num_true=num_true,
                      num_total=num_total))
    lift = pd.DataFrame(r)
    bs_num_true = (cmp.y == 1).sum()
    bs_num_total = cmp.shape[0]
    bs_rprate = bs_num_true / bs_num_total
    lift['baseline'] = bs_rprate
    lift['lift'] = lift.responserate / bs_rprate
    return lift

def plot_lift(ytrue, yproba, cumulative=True):
    import matplotlib.ticker as ticker
    lift = calculate_lift(ytrue, yproba, cumulative)
    grp = 10 - lift.decile*10
    drr = lift.responserate
    rrr = lift.baseline
    fig = plt.figure()
    ax = fig.add_subplot(111)
    line1, = ax.plot(grp, drr, marker='o', color='blue', lw=1.5)
    line2, = ax.plot(grp, rrr, ls='dashed', color='gray', lw=1.5)

    xtext = ax.set_xlabel('Deciles')
    ytext = ax.set_ylabel('% Response')

    ax.set_xlim(0.5, 9.5)
    ax.set_xticks(range(1, 11, 1))

    ax.set_ylim(0, max(drr)*(1.1))
    ax.yaxis.set_major_formatter(
        ticker.FuncFormatter(lambda t1, _: '{:.0%}'.format(t1)))

    txt1 = str("Decile 1 R.R.: " + '{:.1%}'.format(
        drr.loc[0]) + "; Lift: " + '{:3.2}'.format(drr.loc[0]/rrr.loc[0]))
    txt2 = str("Decile 2 R.R.: " + '{:.1%}'.format(
        drr.loc[0]) + "; Lift: " + '{:3.2}'.format(drr.loc[1]/rrr.loc[1]))
    txt3 = str("Decile 3 R.R.: " + '{:.1%}'.format(
        drr.loc[0]) + "; Lift: " + '{:3.2}'.format(drr.loc[2]/rrr.loc[2]))

    plt.text(grp[1]-0.7, drr.loc[0]*0, 99, txt1,
             ha='left', rotation=0, wrap=True)
    plt.text(grp[2]-0.7, drr.loc[0]*0, 99, txt2,
             ha='left', rotation=0, wrap=True)
    plt.text(grp[3]-0.7, drr.loc[0]*0, 99, txt3,
             ha='left', rotation=0, wrap=True)

    plt.legend((line1, lin2), ('Model', 'Baseline'),
               loc='upper right', bbox_to_anchor=[0.95, 0.95], shadow=True)

    if cumulative:
        plt.suptitle('Cumulative Lift Chart', fontsize=14)
    else:
        plt.suptitle('Stratified Lift Chart', fontsize=14)
    plt.show()

def plot_confusion_matrix(ytrue,
                          ypred,
                          classes=None,
                          normalize=True,
                          cmap=None):
    cmap = plt.cm.Blues if not cmap else cmap
    cm = confusion_matrix(ytrue, ypred)
    if not classes:
        n = len(set(ytre))
        classes = list(range(n))
    if normalize:
        title = 'Normalized Confusion Matrix'
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
    else:
        title = 'Frequency Confusion Matrix'
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, format(cm[i, j], fmt),
                 horizontalalignment='center',
                 color='white' if cm[i, j] > thresh else 'black')

    plt.ylabel("True label")
    plt.xlabel('Predicted label')
    plt.tight_layout()


def kfold_cv(model, X, y, cv, verbose=1, shuffle=False):
    from sklearn.metrics import mean_squared_error
    return kfold_cvx(model, X, y, cv, 
            f_metric=mean_squared_error,
            verbose=verbose,
            shuffle=shuffle)

def kfold_cvx(model,
             X,
             y,
             cv,
             f_metric=mean_squared_error,
             verbose=1,
             shuffle=False):
    from copy import deepcopy
    from sklearn.model_selection import KFold
    kf = KFold(n_splits=cv, shuffle=shuffle)
    kfscores = []
    for i, (train_index, test_index) in enumerate(kf.split(X)):
        if verbose:
            print("processing cv batch {} of {}".format(i+1, cv), end=' ')
        mdl = deepcopy(model)
        X_train, X_test = X.loc[train_index], X.loc[test_index]
        y_train, y_test = y[train_index], y[test_index]
        mdl.fit(X_train, y_train)
        
        y_train_pred = mdl.predict(X_train)
        trainscore = f_metric(y_train, y_train_pred)
        
        y_test_pred = mdl.predict(X_test)
        testscore = f_metric(y_test, y_test_pred)
        scores = {'train': trainscore, 'test':testscore}
        print("scores: train {}, test {}".format(trainscore, testscore))
        kfscores.append(scores)
    rst = pd.DataFrame(kfscores)
    stats = rst.describe().loc[['min', 'mean', 'std', 'max']] 
    return pd.concat([rst, stats])

def plot_history(history ,figsize=(4,2)):
    """Plot Keras training history"""
    loss_list = [s for s in history.history.keys() if 'loss' in s and 'val' not in s]
    val_loss_list = [s for s in history.history.keys() if 'loss' in s and 'val' in s]
    acc_list = [s for s in history.history.keys() if 'acc' in s and 'val' not in s]
    val_acc_list = [s for s in history.history.keys() if 'acc' in s and 'val' in s]

    if len(acc_list)==0:
        fig, axloss = plt.subplots(1, figsize=(figsize[0], figsize[1]/2))
        axloss.set_xlabel('Epochs')
    else:
        fig, (axloss, axacc) = plt.subplots(2, sharex=True, figsize=figsize)
        axacc.set_ylim(0,1)

    ## As loss always exists
    epochs = range(1,len(history.history[loss_list[0]]) + 1)
    
    ## Loss
    for l in loss_list:
        axloss.plot(epochs, history.history[l], 'b', label='Training loss (' + str(str(format(history.history[l][-1],'.5f'))+')'))
    for l in val_loss_list:
        axloss.plot(epochs, history.history[l], 'g', label='Validation loss (' + str(str(format(history.history[l][-1],'.5f'))+')'))

    axloss.set_ylabel('Loss')
    axloss.legend(bbox_to_anchor=(1.04,1), loc="upper left")
    
    if len(acc_list)>0:
        ## Accuracy
        for l in acc_list:
            axacc.plot(epochs, history.history[l], 'b', label='Training accuracy (' + str(format(history.history[l][-1],'.5f'))+')')
        for l in val_acc_list:    
            axacc.plot(epochs, history.history[l], 'g', label='Validation accuracy (' + str(format(history.history[l][-1],'.5f'))+')')

        axacc.set_xlabel('Epochs')
        axacc.ylabel('Accuracy')
        axacc.legend(bbox_to_anchor=(1.04,1), loc="upper left")



from keras.layers import Input, Dense
from keras.models import Model, Sequential
import numpy as np
def get_AE(units):
    """ build an autoencoder with explict encoder and decorder object

    Parameters
    =========
    units : list of integers
        The number of units (neurons) in each layer as in the following order:
        [n_inputs, h1, h2, ..., middle_layer]. See examples below.
    
    Returns
    =======
    autoencoder : keras Model
        The autoencoder model, made of encoder+decoder.
    
    encoder : keras Model
        The encoder part of the autoencoder.
    
    decoder : keras model
        The decoder part of the autoencoder.
    
    Examples
    ========
    Given parameter units = [10, 5, 3], this represents an autoencoder of structure [10, 5, 3, 5, 10].
    Specifically, it has 10 inputs, 5 units in the first hidden layer, 3 units in the second
    hidden layer (the bottleneck) and 5 units in the decoder layer, and 10 units in the output
    layer (same as the number of inputs).
    """
    n_encoder_inputs = units[0]
    n_encoder_units = units[1:]
    n_decoder_inputs = units[-1]
    n_decoder_units = units[:-1][::-1]
    # build encoder
    encoder = Sequential()
    encoder.add(Dense(n_encoder_units[0], input_dim=n_encoder_inputs, activation='sigmoid'))
    for n in n_encoder_units[1:]:
        encoder.add(Dense(n, activation='sigmoid'))
    # build decoder
    decoder = Sequential()
    decoder.add(Dense(n_decoder_units[0], input_dim=n_decoder_inputs,activation='sigmoid'))
    for n in n_decoder_units[1:]:
        decoder.add(Dense(n, activation='sigmoid'))
    # build the autoencoder
    autoencoder = Sequential(encoder.layers + decoder.layers)
    return (autoencoder, encoder, decoder)

def add_noise_masking(noiseParam, data):
    """Add noise to the input data by setting a proportion of each observation to 0.
    This function adds noise as described in Vincent et al. (2008). 
    https://www.iro.umontreal.ca/~vincentp/Publications/denoising_autoencoders_tr1316.pdf
    Parameters
    ==========
    noiseParam : float, in range [0,1]
        The proportion of inputs to be set to 0 to introduce noise. Particularly,
        0 means adding no noise, 1 means setting all inputs to 0 (not useful).
    
    data : 2D matrix, each row is an observation of X
        So, data(i,:) is the i-th training example.
    
    Returns
    =======
    noisyData : 2D matrix, each row is an observation of X_tild
        Original data with noise.
    """
    noiseMask = np.ones(data.shape)
    n_featuers = data.shape[1]
    n_noised = int(n_featuers * noiseParam)
    
    # add noise by setting part of the mask to zero
    noise_vector = np.ones(n_featuers)
    noise_vector[:n_noised] = 0

    # shuffle the noise in each row
    for i in range(noiseMask.shape[0]):
        np.random.shuffle(noise_vector)
        noiseMask[i,:] = noise_vector
    
    noisyData = data * noiseMask
    return noisyData

def get_NN(units):
    model = Sequential()
    n_features = units[0]
    for i, n in enumerate(units[1:]):
        if i==0:
            model.add(Dense(units=n, activation='sigmoid', input_dim=n_features))
        else:
            model.add(Dense(units=n, activation='sigmoid'))
    return model


import pandas as pd
from sklearn.preprocessing import MinMaxScaler, PolynomialFeatures
from sklearn.base import BaseEstimator, TransformerMixin
from .utilities import plot_history
import sklearn.pipeline

def set_debug(steps, flag=True, level=0):
	""" recursively set debugging (verbose) status of pipeline items
	"""
	for (name, step) in steps:
		if isinstance(step, (iPipeline, sklearn.pipeline.Pipeline)):
			# set_debug(step.steps, flag, level=level+1)
			# indent is disabled
			set_debug(step.steps, flag, level=0)
		else:
			if flag:
				if not hasattr(step, '_backup_fit'):
					step._backup_fit = step.fit
				else:
					step.fit = step._backup_fit	
				step._DEBUG_ = True
				step._DESC_ = '\t'*level + name
				def decorate_fit(func):
					def wrapper(self, *args, **kwargs):
						_name = self.__class__.__name__
						print("\t"*level, _name, "is fitting..", end=' ')
						f = func(*args, **kwargs)
						print("Ok")
						return f
					return wrapper
				functype = type(step.fit)
				step.fit = functype(decorate_fit(step.fit), step)

				if hasattr(step, 'transform'):
					if not hasattr(step, '_backup_transform'):
						step._backup_transform = step.transform
					else:
						step.transform = step._backup_transform	
					def decorate_transform(func, level):
						def wrapper(self, *args, **kwargs):
							_name = self.__class__.__name__
							print("\t"*level, _name, end=" => ")
							r = func(*args, **kwargs)
							print(r.shape)
							return r
						return wrapper
					functype = type(step.transform)
					step.transform = functype(decorate_transform(step.transform, level), step)
			else:
				if hasattr(step, '_backup_fit'):
					step.fit = step._backup_fit	
				if hasattr(step, 'transform') and hasattr(step, '_backup_transform'):
					step.transform = step._backup_transform
				
class iPipeline(sklearn.pipeline.Pipeline):
	_DEBUG_ = False
	_DESC_ = 'iPipeline'
	def __init__(self, steps, memory=None):
		super().__init__(steps, memory)
	def set_verbose(self, value):
		set_debug([('iPipeline', self)], value)

def make_pipeline(*steps, **kwargs):
	"""Construct a customized Pipeline from the given estimators.
	"""
	memory = kwargs.pop('memory', None)
	if kwargs:
		raise TypeError('Unknown keyword arguments: "{}"'
						.format(list(kwargs.keys())[0]))
	return iPipeline(sklearn.pipeline._name_estimators(steps), memory=memory)

class NoFitTransformer(BaseEstimator, TransformerMixin):
	'''A base class for transformers that does not require fit()'''
	def __init__(self):
		pass        
	def fit(self, X, y=None):
		return self

class BaseTransformer(BaseEstimator, TransformerMixin):
	'''A base class for dataframe transformation. It implements a 
	column list parameter for you so you don't have to repeat yourself.'''
	_DEBUG_ = False
	_DESC_ = 'BaseTransformer'
	def __init__(self, cols=None):
		self.cols = cols
	def get_cols(self, X, cols):
		columns = X.columns.tolist()
		sel = []
		if isinstance(cols, str):
			if cols == '__all__':
				sel = columns
			else:
				sel = [cols]
		else:
			sel = cols
		# check avaiability:
		r = []
		for li in sel:
			if li not in columns:
				print("Warning: column '{}' not found in dataset.".format(li))
			else:
				r.append(li)
		return r
	def fit(self, X, y=None):
		r = self._fit(X, y)
		# if self._DEBUG_:
		# 	print("{} is fitting..".format(self._DESC_))
		# 	self._debug_fit()
		return r
	def _fit(self, X, y):
		return self
	def transform(self, X):
		r = self._transform(X)
		# if self._DEBUG_:
		# 	print("{} : output data shape {}".format(self._DESC_, r.shape))
		# 	self._debug_transform()
		return r
	def _debug_fit(self):
		pass
	def _debug_transform(self):
		pass
	def _transform(self, X):
		return X
	def summary(self, X):
		msg = 'summary message'
		print(msg)


class Select(BaseTransformer):
	"""Select columns by name"""
	def _transform(self, X):
		return X.loc[:, self.cols]

class Drop(BaseTransformer):
	def _transform(self, X):
		return X.drop(self.get_cols(X,self.cols), axis=1)
		
class Dropna(BaseTransformer):
	"""Drop observations if the value of any of the columns are missing"""
	def _transform(self, X):
		return X.dropna(axis=0, how='any', subset=self.cols)

class Impute(BaseTransformer):
	"""Inpute columns with the given imputer"""
	def __init__(self, cols=None, value=0):
		self.cols = cols 
		self.value = value
	def _transform(self, X):
		cols = self.cols
		if not cols:
			cols = X.columns.tolist()
		X.loc[:,cols] = X[cols].fillna(self.value)
		return X

class Rescale(BaseTransformer):
	def __init__(self, cols=None, feature_range=(0,1)):
		self.cols = cols
		self.scaler = MinMaxScaler(feature_range=feature_range)
	def _fit(self, X, y=None):
		if not self.cols:
			self.cols = X.columns.tolist()
		self.scaler.fit(X[self.cols])
		return self
	def _transform(self, X):
		X.loc[:, self.cols] = self.scaler.transform(X[self.cols])
		return X

class Normalize():
	def __init__(self, cols):
		self.cols = cols
	def _fit(self, X, y=None):
		self.mean_std = X[self.cols].describe().loc[['mean', 'std']].to_dict()
		return self
	def _transform(self, X):
		for var, stats in self.mean_std.items():
			X[var] = (X[var] - stats['mean']) / stats['std']
		return X

class Interact(BaseTransformer):
	def __init__(self, cols, degree, drop_cols=True, interaction_only=False, normalize=True):
		self.cols = cols
		self.degree = degree
		self.drop_cols = drop_cols
		self.interaction_only = interaction_only
	def _transform(self, X):
		polyer = PolynomialFeatures(self.degree, interaction_only=self.interaction_only, include_bias=False)
		polys = polyer.fit_transform(X[self.cols])
		poly_names = polyer.get_feature_names(self.cols)
		polydf = pd.DataFrame(polys, columns=poly_names)
		if self.drop_cols:
			X = X.drop(self.cols, axis=1)
		X = pd.concat([X.reset_index(drop=True), polydf], axis=1)
		return X

class EncodeDummy(BaseTransformer):
	def __init__(self, cols):
		self.cols = cols
	def _fit(self, X, y=None):
		self.value_dict = dict()
		for li in self.cols:
			self.value_dict[li] = X[li].unique()
		return self
	def _transform(self, X):
		r_df = []
		for li in self.cols:
			for value in X[li].unique():
				if value not in self.value_dict[li]:
					# TODO: show the % of undocumented values?
					print("Warning: feature '{}' has undocumented value '{}'".format(li, value))
			categorical = pd.Categorical(X.loc[:,li], categories=self.value_dict[li])
			tdf = pd.get_dummies(categorical, prefix=li, dummy_na=True, drop_first=True)
			r_df.append(tdf)
			X.drop(li, axis=1, inplace=True)
		X = pd.concat([X.reset_index(drop=True), *r_df], axis=1)
		return X

class Model(TransformerMixin):
	def __init__(self, model, output='class'):
		self.model = model
		self.output = output

	def fit(self, *args, **kwargs):
		self.model.fit(*args, **kwargs)
		return self

	def transform(self, X, **transform_params):
		if self.output == 'class':
			return self.model.predict(X)
		elif self.output == 'proba':
			return self.model.predict_proba(X)
		else:
			raise ValueError(
				"Unrecognized output format : {}. Choose 'class' or 'proba'.".format(self.output))

class Describe(BaseTransformer):
	def _transform(self, X):
		return X

class Tokenize(BaseTransformer):
	def __init__(self, tokenizer, maxlen):
		self.tokenizer = tokenizer
		self.maxlen=maxlen
	def _fit(self, X, y=None):
		self.tokenizer.fit_on_texts(X)
		return self
	def _transform(self, X):
		from keras.preprocessing import sequence
		seq = self.tokenizer.texts_to_sequences(X)
		seqmat = sequence.pad_sequences(seq, maxlen=self.maxlen)
		return seqmat

class Autoencoder(BaseTransformer):
	def __init__(self, units, optimizer='RMSprop', loss='mean_squared_error',
				noise_param=0.2,
				batch_size=512,
				epochs=400,
				verbose=0):
		self.units = units
		self.optimizer = optimizer
		self.loss = loss
		self.noise_param = noise_param
		self.batch_size = batch_size
		self.epochs = epochs
		self.verbose = verbose
	def _fit(self, X, y=None):
		from .engines import get_AE, add_noise_masking
		m = X.shape[1]
		self.autoencoder, self.encoder, self.decoder = get_AE([m] + self.units)
		self.autoencoder.compile(optimizer=self.optimizer, loss=self.loss)

		Xnoised = add_noise_masking(self.noise_param, X)
		self.history = self.autoencoder.fit(Xnoised,X, 
				batch_size=self.batch_size, 
				epochs=self.epochs,
				verbose=self.verbose)
		return self
	def _debug_fit(self):
		return plot_history(self.history)
	def _transform(self, X):
		return self.encoder.predict(X)

class NN(BaseTransformer):
	def __init__(self, units, optimizer='RMSprop', loss='mean_squared_error',
				batch_size=512,
				epochs=400,
				metrics=None,
				verbose=0):
		self.units = units
		self.optimizer = optimizer
		self.loss = loss
		self.batch_size = batch_size
		self.epochs = epochs
		self.metrics = metrics
		self.verbose = verbose
	def _fit(self, X, y=None):
		from .engines import get_NN
		m = X.shape[1]
		self.nn = get_NN([m] + self.units)
		self.nn.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics)
		self.history = self.nn.fit(X, y, 
				batch_size=self.batch_size, 
				epochs=self.epochs,
				verbose=self.verbose)
		return self
	def _debug_fit(self):
		return plot_history(self.history)
	def _transform(self, X):
		return self.nn.predict(X)
	def score(self, X, y):
		from sklearn.metrics import accuracy_score
		return accuracy_score(y, self.nn.predict_classes(X))
	def predict(self, X):
		return self.nn.predict_classes(X)

class XGB(BaseTransformer):
	def __init__(self, output='class', n_estimators=100, learning_rate=0.1, n_jobs=4, max_depth=2, optimizer='count:poisson'):
		import xgboost as xgb
		self.output = output
		self.model = xgb.XGBClassifier(n_estimators=n_estimators,
								learning_rate=learning_rate,
								n_jobs=n_jobs,
								max_depth=max_depth,
								optimizer=optimizer
								)
	def _fit(self, X, y):
		self.model.fit(X,y)
		return self
	def _transform(self, X):
		if self.output=='class':
			r = self.model.predict(X)
		elif self.output=='proba':
			r = self.model.predict_proba(X)
		else:
			raise ValueError("Unrecognized output format : {}. Choose 'class' or 'proba'.".format(self.predict))
		return r



from __future__ import print_function
import pickle
import time
import os

import logging
from logging import handlers
import sys


def cprint(*args, **kwargs):
    """Markup words wrapped in << and >> and print them out.
    
    Parameters
    =========
    b : bool, default True
        Use bold font.
    c : str, default 'yellow'
        Color of the font.
    h : bool, default Flase
        Highlight the font.
    """
    b = kwargs.pop('b', True)
    c = kwargs.pop('c', 'yellow')
    h = kwargs.pop('h', False)
    color2num = dict(
        gray=30,
        red=31,
        green=32,
        yellow=33,
        blue=34,
        magenta=35,
        cyan=36,
        white=37,
        crimson=38
    )
    # Import six here so that `utils` has no import-time dependencies.
    # We want this since we use `utils` during our import-time sanity checks
    # that verify that our dependencies (including six) are actually present.
    import six
    attr = []
    num = color2num[c]
    if h:
        num += 10
    attr.append(six.u(str(num)))
    if b:
        attr.append(six.u('1'))
    attrs = six.u(';').join(attr)
    prefix = six.u('\x1b[%sm' % attrs)
    suffix = six.u('\x1b[0m')

    s = args[0]
    s = s.replace('<<', prefix)
    s = s.replace('>>', suffix)
    args = (s, *args[1:])

    return print(*args, **kwargs)

# from OpenAi.gym
def colorize(string, color, bold=False, highlight=False):
    """Return string surrounded by appropriate terminal color codes to
    print colorized text. Valid colors: gray, red, green, yellow,
    blue, magenta, cyan, white, crimson
    """
    color2num = dict(
        gray=30,
        red=31,
        green=32,
        yellow=33,
        blue=34,
        magenta=35,
        cyan=36,
        white=37,
        crimson=38
    )
    # Import six here so that `utils` has no import-time dependencies.
    # We want this since we use `utils` during our import-time sanity checks
    # that verify that our dependencies (including six) are actually present.
    import six
    attr = []
    num = color2num[color]
    if highlight: num += 10
    attr.append(six.u(str(num)))
    if bold: attr.append(six.u('1'))
    attrs = six.u(';').join(attr)
    return six.u('\x1b[%sm%s\x1b[0m') % (attrs, string)

from pathlib import Path
class FileIO():
    @staticmethod
    def glob(pattern):
        '''Get a list of files that satisfy the given pattern. '''
        from glob import glob
        return glob(pattern, recursive=True)
    @staticmethod
    def exists(path):
        return Path(path).exists()
    @staticmethod
    def is_dir(path):
        return Path(path).is_dir()
    @staticmethod
    def is_file(path):
        return Path(path).is_file()
    @staticmethod
    def stat(path):
        return Path(path).stat()
    @staticmethod
    def get_folder(path):
        return Path(path).parent
    @staticmethod
    def resolve(path):
        """Resolve a relative path to an absolute path (e.g., ../file.txt -> D:/data/file.txt)."""
        return str(Path(path).resolve())
    @staticmethod
    def get_ext(path):
        "Return the extension of the file (e.g., C:/a/b/c.txt -> .txt)."
        return Path(path).suffix
    @staticmethod
    def with_ext(path, ext):
        "Return a path with file extension replaced (e.g., C:/a/b/c.txt -> C:a/b/c.log)."
        return str(Path(path).with_suffix(ext))
    @staticmethod
    def filename(path):
        "Return the filename of path (e.g., C:/a/b/c.txt -> c.txt)."
        return Path(path).name
    @staticmethod
    def makedirs(dirpath):
        '''Create (sub)directories in the path. Do nothing if exists.'''
        try:
            os.makedirs(dirpath)
        except OSError:
            pass
    @staticmethod
    def validate_filename(s):
        """
        Return the given string converted to a string that can be used for a clean
        filename. Remove leading and trailing spaces; convert other invalid chars to
        underscores; and remove anything that is not a ASCII char.

        Args:
            s (string): the file name, NOT the path of the file.
        
        Returns:
            string: validated file name for Windows system
        """
        s=s.strip()
        invalid_chars=['"', '\\', "'", ".", "/", " ", "?", ":", "<", ">", "|", "*"]
        for char in invalid_chars:
            s=s.replace(char,'_')
        s = ''.join((c for c in s if 0 < ord(c) < 127)) # remove non-ASCII chars
        return s

def getFileList(*args, **kwargs):
    raise ValueError("getFileList function is deprecated. Please use ytZoo.FileIO.glob instead.")

def set_show_all_data():
    """Enable Jupyter Notebook to display multiple outputs in cells."""
    from IPython.core.interactiveshell import InteractiveShell
    InteractiveShell.ast_node_interactivity = 'all'

def multiprocess(data, worker, cores=4):
    ''' map {worker} function to each of {data}, return the aggregated result'''
    import concurrent.futures
    r = []
    with concurrent.futures.ProcessPoolExecutor(max_workers=cores) as executor:
        for inputdata, result in zip(data, executor.map(worker, data)):
            if isinstance(result,list):
                r.extend(result)
            else:
                r.append(result)
    return r

def chunk_process_and_save(data, worker, outfilename, step=100, cores=4):
    '''process large amounts of {data} chunk by chunk using the {worker},save step results to disk'''
    from tqdm import tqdm
    def chunkit(seq, n):
        for i in range(0, len(seq), n):
            yield seq[i:i + n]
    total = len(data)
    for i,part in tqdm(list(enumerate(chunkit(data, step))),ascii=True):
        foutname = "{}_{}to{}of{}.pkl".format(outfilename,i*step,(i+1)*step,total)
        r = multiprocess(part, worker, cores)
        pkl.dump(r,foutname)
    return True

# Sequence operations #############################################    
def chunk_even(seq, num):
    'Split a sequence into evenly sized chunks'
    avg = len(seq) / float(num)
    out = []
    last = 0.0
    while last < len(seq):
        out.append(seq[int(last):int(last + avg)])
        last += avg
    return out

def chunk_size(seq, n):
    'Yield successive n-sized chunks from seq.'
    for i in range(0, len(seq), n):
        yield seq[i:i + n]

from itertools import islice
def moving_window(seq, n):
    "Returns a moving window (of width n) over data from the iterable"
    "   s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...                   "
    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result

###########################################################################

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

def today_date():
    import datetime
    return str(datetime.date.today())
        
def cmd(commands):
    from subprocess import call
    return call(commands)

class pkl:
    @staticmethod
    def load(fname):
        with open(fname,'rb') as fin:
            return pickle.load(fin)
    @staticmethod
    def dump(data,fname):
        with open(fname,'wb') as fout:
            pickle.dump(data,fout)

class Dumper(object):
	"""Pickle a list and dump it to disk every {max_length} items."""
	def __init__(self, path='output.pkl', max_length=1000):
		self.path = path
		self.buffer = []
		self.max_length = max_length
		self._count = 0
	def __enter__(self):
		return self
	def _get_output_path(self):
		basepath, ext = os.path.splitext(self.path)
		path = "{}_{}{}".format(basepath, self._count, ext)
		return path
	def append(self, x):
		self.buffer.append(x)
		if len(self.buffer) >= self.max_length:
			self.flush()
	def extend(self, alist):
		self.buffer.extend(alist)
		if len(self.buffer) >= self.max_length:
			self.flush()
	def flush(self):
		with open(self._get_output_path(), 'wb') as fout:
			pickle.dump(self.buffer, fout)
		self.buffer = []
		self._count += 1
	def __exit__(self, type, value, traceback):
		if len(self.buffer) > 0:
			self.flush()


def logger(loggerName=None, 
            logfile=None, 
            stdout=True, 
            level_stdout='INFO',
            level_file='DEBUG',
            date='short' 
            ):
    """Return a logger that can write to stdout and a file. Replace the logger with the same name if exists.
    
    Parameters
    ----------
    loggerName : string, optional
        name of the logger. Pass None to use the module name.
    logfile : string, optional
        a file path indicating the file to write the logs. Pass None to disable writing to file. Default is None.
    stdout : bool, optional
        whether enable writing to standard output. Default is true.
    level_stdout : str, default 'INFO'
        logging level to stdout.
    level_stdout : str, default 'DEBUG'
        logging level to file.
    date : str, default 'short'
        the length of date for each line of log. Choose from None - disable leading dates, 'short' - only time, 
        'long' - date and time.
    Returns
    -------
    logger
        a logger from the built-in logging library
    """
    rootLogger = logging.getLogger(loggerName or __name__)
    # set global level to the lowest
    rootLogger.setLevel(logging.DEBUG)
    for handler in rootLogger.handlers:
        handler.flush()
        handler.close()
        rootLogger.removeHandler(handler)
    rootLogger.handlers = []

    if date == 'short':
        datefmt = "%H:%M:%S"
    elif date == 'long':
        datefmt = "%Y-%m-%d %H:%M:%S"
    elif date is None:
        datefmt = ''
    else:
        raise ValueError("date can only be None, 'short', or 'long'. Got {}.".format(date))
    logFormatter = logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s",datefmt)
    # make sure we don't have duplicated handlers
    has_console_handler = has_file_handler = False
    for handler in rootLogger.handlers:
        if isinstance(handler, logging.handlers.TimedRotatingFileHandler):
            has_file_handler = True
            continue
        if isinstance(handler, logging.StreamHandler):
            has_console_handler = True
            continue
    # add a console handler if non-exists
    if stdout == True and not has_console_handler:
        consoleHandler = logging.StreamHandler(stream=sys.stdout)
        consoleHandler.setLevel(level_stdout)
        consoleHandler.setFormatter(logFormatter)
        rootLogger.addHandler(consoleHandler)
    # add file handler if non-exists
    if logfile!=None and not has_file_handler:
        fileHandler = logging.handlers.TimedRotatingFileHandler(logfile,
                                                                when='midnight',
                                                                backupCount=7)
        fileHandler.setLevel(level_file)
        fileHandler.setFormatter(logFormatter)
        rootLogger.addHandler(fileHandler)
    return rootLogger
    
class Decorators:
    '''A collection of function decorators'''
    @staticmethod
    def cacher(f):
        '''cache the result of a function'''
        cache = {}
        def helper(x):
            if x not in cache:            
                cache[x] = f(x)
            return cache[x]
        return helper

    @staticmethod
    def tracer(f):
        '''trace the call route of a funciton'''
        def helper(*args, **kwargs):
            params=", ".join([str(li) for li in args])+\
                " ".join(",{0}={1}".format(k,kwargs[k]) for k in kwargs)
            call_str = "{0}({1})".format(f.__name__, params)
            print("Calling {0} ...".format(call_str))
            result = f(*args,**kwargs)
            print("... returning from {0} = {1}".format(
                  call_str, result))
            return result
        return helper

    @staticmethod
    def timer(f):
        '''calculate the execution time of the a function'''
        def helper(*args, **kwargs):
            params=", ".join([str(li) for li in args])+\
                " ".join(",{0}={1}".format(k,kwargs[k]) for k in kwargs)
            call_str = "{0}({1})".format(f.__name__, params)
            start_time=time.time()
            result=f(*args,**kwargs)
            milsec=int((time.time()-start_time)*1000)
            print("{0} finished in {1}ms.".format(call_str, milsec))
            return result
        return helper

def follow(thefile):
    '''follows a text file indefinitely.'''
    thefile.seek(0,2)
    while True:
        line = thefile.readline()
        if not line:
            time.sleep(0.1)
            continue
        yield line

def WriteFile(src,dst,overwrite=False):
    FileIO.makedirs(os.path.dirname(dst))
    if os.path.isfile(dst) and overwrite==False:
        raise ValueError("File exists: {}".format(dst))
    with open(dst, 'wb') as f:
        f.write(src)
writeFile=WriteFile

def humanize_filesize(num, suffix='B'):
    '''Translate file size in bytes to human-readable representaions.'''
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)

def xenumerate(sequence, start=1, as_pct=False):
    '''Returns a string in the format of "n/total", instead of just n. 
    This indicates the current relative position of iteration.'''
    n = start
    total=len(sequence)
    for elem in sequence:
        if not as_pct:
            s="%s/%s" % (n,total)
        else:
            s="{0:.2%}".format(n/total)
        yield s, elem
        n += 1

class ExceptionHook(object):
    instance = None
    def __call__(self, *args, **kwargs):
        if self.instance is None:
            from IPython.core import ultratb
            self.instance = ultratb.VerboseTB(call_pdb=True)
        return self.instance(*args, **kwargs)
def install_exception_hook():
    import sys
    sys.excepthook=ExceptionHook()


class using():
    """Backup the current globals(), populate it with classes and functions with in a module, and 
    restores the backup afterwards. Note that all objects created after the backup with be lost.

    Example
    =======
    import ytZoo
    with using('ytZoo'):
        soup = get_soup("https://www.google.com")
    soup
    """
    def __init__(self, module):
        self.module = module
    def __enter__(self):
        self.GLOBALS_BACKUP = globals().copy()
        self.module_dict = self.GLOBALS_BACKUP[self.module].__dict__
        globals().update(self.module_dict)
    def __exit__(self, *args, **kwargs):
        for li in self.module_dict:
            globals().pop(li)
        globals().update(self.GLOBALS_BACKUP)



'''
A collection of data science utilities.
Most of the class and functions are Pandas related.
'''
from __future__ import print_function

from .pandas_api import register_dataframe_booster, register_series_booster

from functools import wraps
from ytZoo import xenumerate
import os
import re
import sys, math
import numpy as np
import pandas as pd
import sqlite3

class SQL:
    "An database interface that integrated Dataframe and SQL."
    def __init__(self, database):
        self._conn = sqlite3.connect(database, detect_types=sqlite3.PARSE_DECLTYPES)
        self._conn.row_factory = sqlite3.Row

    def read_query(self, sql, index_col=None, coerce_float=True, params=None, parse_dates=None, chunksize=None):
        """Ported API from pd.read_sql_query."""
        return pd.read_sql_query(sql=sql,
                                con=self._conn,
                                index_col=index_col,
                                coerce_float=coerce_float,
                                params=params,
                                parse_dates=parse_dates,
                                chunksize=chunksize)
    def execute(self, sql, parameters=None):
        "Ported from sqlite3.Connection.execute()."
        return self._conn.execute(sql)
    
    def commit(self):
        "Commit changes."
        self._conn.commit()

    def executemany(self, sql, parameters, commit=True):
        """Ported from sqlite3.Connection.executemany().

        Parameters
        ==========
        commit : bool, default True
            Commit changes after executing the command.
        """
        status = self._conn.executemany(sql, parameters)
        if commit:
            self.commit()
        return status

    def close(self):
        self._conn.close()
    def tables(self):
        "Return a list of table names in the database."
        return self.read_query("""SELECT tbl_name FROM sqlite_master
                                WHERE type=='table'""")['tbl_name'].tolist()
    def columns(self, tbl_name):
        "Return a list of column names in a table."
        assert ' ' not in tbl_name, "Invalid table name '{}'".format(tbl_name)
        cursor = con.execute("select * from {} limit 1".format(tbl_name))
        names = list(map(lambda x: x[0], cursor.description))
        return names
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        self.close()

@register_dataframe_booster('merge')
@wraps(pd.merge)
def check_merge(left, *args, **kwargs):
    """Merge datasets and show statistics."""
    kwargs['indicator'] = True
    mg = left.merge(*args, **kwargs)
    # report stats
    stats = mg['_merge'].value_counts()
    stats.name = 'obs.'
    stats = stats.pipe(cum_freq)
    try:
        from IPython.core.display import display, HTML
        display(HTML(stats.to_html()))
    except ImportError:
        print(stats)
    return mg.drop('_merge', axis=1)

@register_dataframe_booster('update')
def check_update(master, withdf, match, columns, append_new=False):
    """Update columns in the master table with the updating table.
    Match records using 'match' as the key.

    Parameters
    ==========
    master : DataFrame
    withdf : DataFrame
        Use data in this dataframe to update the master.
    match : str
        The merging key.
    columns : [str,]
        Columns to update. Must exist be master and withdf under the same names.
    append_new : bool, default False
        Whether to append new IDs (match) from withdf but not in master, in the result.
        In other words, default append_new=False will not change IDs in master, but only
        update new values of them in withdf.
    """
    assert isinstance(
        columns, list), "columns must be a list, got {}.".format(columns)
    assert isinstance(
        match,   str), "match must be a string, got {}.".format(match)
    mastercols = master.columns
    withdfcols = withdf.columns
    for col in columns+[match]:
        assert col in mastercols, "Column '{}' not in master table.".format(col)
        assert col in withdfcols, "Column '{}' not in updating table.".format(col)

    # discard irrelevant columns in the updating table
    withdf = withdf[columns+[match]]
    # suffixes to distinguish master table columns from updating table columns
    sfx, sfy = suffixes = ('_x', '_y')
    how = 'outer' if append_new else 'left'
    mg = check_merge(master, withdf, how=how, validate='1:1',
                     on=match, suffixes=suffixes)

    # update values
    for col in columns:
        mask = mg[col+sfy].notnull()
        mg.loc[mask, col+sfx] = mg.loc[mask, col+sfy]
        mg = mg.drop(col+sfy, 1)
        mg = mg.rename(columns={col+sfx: col})
    return mg

@register_dataframe_booster('viewpct')
def pct(dataframe, axis=0):
    """Transform values in each cell to percentages repecting to
    the axis selected.

    Example:
    df = 
        Row1 1 2 3
        Row2 1 2 3
    >>> df.pipe(pct, 1)
    Out:
        Row1 50% 50% 50%
        Row2 50% 50% 50%
        Total 100% 100% 100%
    """
    if axis == 0:
        prepare = dataframe.pipe(sumup, 0, 1)
        prepare = prepare.div(prepare['Total'], axis=0)
    if axis == 1:
        prepare = dataframe.pipe(sumup, 1, 0)
        prepare = prepare.div(prepare.loc['Total', :], axis=1)
    prepare = prepare.applymap(lambda x: "{:.0%}".format(x))
    return prepare


def format_percentage(n, precision='auto'):
    """Display a decimal number in percentage.
    
    Parameters
    =========
    n : float
        The number to format.
    percision: int, str, default 'auto'
        The precision of outcome. Default 'auto' to automatically
        choose the least precision on which the outcome is not zero.

    Examples
    ========
    format_percentage(0.001) ==> '0.1%'
    format_percentage(-0.0000010009) ==> '-0.0001%'
    format_percentage(0.001, 4) ==> '0.1000%'
    """
    if precision == 'auto':
        if n != 0:
            k = abs(n)*100
            cnt = 0
            while k < 1:
                k = k * 10
                cnt += 1
            precision = cnt
        else:
            precision = 0
    fmt = '{:.'+str(precision)+'%}'
    return fmt.format(n)

@register_dataframe_booster('fmt')
def format_dataframe(dataframe, big=0, pct='auto'):
    """Automatically format a dataframe with business readings and percentages.
    
    Parameters
    ==========
    dataframe : pd.DataFrame
    big : int, default 0
        Precision for big numbers.
    pct : int, default 'auto'
        Precision for percentage numbers.
    """
    df = dataframe.copy()

    def is_pct(srs):
        if ('pct' in srs.name.lower()) or ('%' in srs.name) \
            or ('percent' in srs.name.lower()):
            return True
        return False
    allcols = set(df.columns)
    pctcols = {col for col in allcols if is_pct(df[col])}
    numcols = allcols - pctcols
    pctcols = list(pctcols)
    numcols = list(numcols)
    df[pctcols] = df[pctcols].applymap(lambda x: format_percentage(x, pct))
    df[numcols] = df[numcols].applymap(lambda x: bignum(x, big))
    return df

@register_dataframe_booster('markup')
def markup(dataframe, precision=0):
    """ Apply human readble formats and add total column and row """
    if isinstance(dataframe, pd.Series):
        dataframe = dataframe.to_frame()
    return dataframe.pipe(sumup, 1, 1).applymap(lambda x: bignum(x, precision))

@register_dataframe_booster('sumup')
def sumup(dataframe, row=True, column=False):
    """ Add a total row and/or a total column"""
    if row:
        rowtot = dataframe.sum(0).T
        if isinstance(dataframe.index, pd.core.index.MultiIndex):
            row_idx_lvls = len(dataframe.index.levels)
            tot_row_name = [''] * row_idx_lvls
            tot_row_name[0] = 'Total'
            rowtot.name = tuple(tot_row_name)
        else:
            rowtot.name = 'Total'
        dataframe = dataframe.append(rowtot, ignore_index=False, sort=True)
        # dataframe = pd.concat([dataframe, rowtot], axis=0, sort=False)
    if column:
        coltot = dataframe.sum(1)
        coltot.name = 'Total'
        dataframe['Total'] = coltot
        # dataframe = pd.concat([dataframe, coltot], axis=1, sort=False)
    return dataframe

def bignum(n, precision=0):
    """ Transform a big number into a business style representation.
    
    Example:
    >>> bignum(123456)
    Output: 123K
    """
    millnames = ['', 'K', 'M', 'B', 'T']
    try:
        n = float(n)
        millidx = max(0, min(len(millnames)-1,
                             int(math.floor(0 if n == 0 else math.log10(abs(n))/3))))
        fmt = '{:.'+str(precision)+'f} {}'
        return fmt.format(n / 10**(3 * millidx), millnames[millidx])
    except ValueError:
        return n

@register_dataframe_booster('nmissing')
def nmissing(dataframe, show_all=False):
    """ Evaluate the number of missing values in columns in the dataframe """
    total = dataframe.shape[0]
    missing = pd.isnull(dataframe).sum().to_frame(name='nmissing')
    missing['pctmissing'] = (missing.nmissing / total * 100).apply(int) / 100
    missing.sort_values(by='nmissing', inplace=True)
    missing.reset_index(inplace=True)
    if not show_all:
        missing = missing.loc[missing.nmissing > 0]
    return missing

@register_dataframe_booster('nlevels')
def nlevels(dataframe, show_values=True):
    """Report the number of unique values (levels) for each variable. 
    Useful to inspect categorical variables.
    """
    nlvl = dataframe.nunique()
    cnt = dataframe.count().to_frame('obs')
    dtype = dataframe.dtypes.to_frame('dtype')
    levels = nlvl.to_frame('levels').join([cnt, dtype])
    levels.loc[levels.obs==0, 'obs'] = np.nan
    if show_values:
        r = []
        for li in dataframe.columns:
            unique_values = dataframe[li].unique()
            sample = ', '.join(map(str, unique_values[:5]))
            if len(unique_values) > 5:
                sample += ', ...'
            r.append(sample)
        levels['values'] = r
    levels.loc[:, 'uniqueness'] = (levels.levels / levels.obs)
    levels = levels.sort_values('uniqueness', ascending=False)
    return levels[['levels', 'obs', 'dtype', 'uniqueness', 'values']]

@register_dataframe_booster('findcolumns')
def find_columns(df, pat, case=False):
    "Filter column names in a dataframe."
    flag = re.IGNORECASE if not case else 0
    return [s for s in df.columns if re.search(pat, s, flag) is not None]


### Series boosters
@register_series_booster('cum')
def cum_freq(series):
    """Return a dataframe showing the percentage, cumulative percentage, 
    and cumulative frequency of a series.
    """
    series.name='freq'
    df = pd.DataFrame(series)
    df['pct'] = (series / series.sum()).round(3)
    df['cumfreq'] = series.cumsum()
    df['cumpct'] = (df.cumfreq / series.sum()).round(3)
    return df
    
@register_series_booster('distplot')
def distplot(srs, normalize=True, color='blue', figsize=(12, 6), grid=False, **kwargs):
    """Plot clear distribution with information about nobs and #missing""" 
    assert srs.dtype == 'float' or srs.dtype == 'int'
    name = srs.name
    ax = srs.hist(grid=grid, color=color, figsize=figsize, **kwargs)
    if normalize:
        ytks = [t/srs.shape[0] for t in ax.get_yticks()]
        ytks = ['{:.0%}'.format(t) for t in ytks]
        ax.set_yticklabels(ytks)
    ax.set_title('{} (ms {:.1%}, {} / {})'.format(
                    name,
                    pd.isnull(srs).sum()/srs.shape[0],
                    pd.isnull(srs).sum(),
                    srs.shape[0]))
    return ax


### Other utilities
def df_to_sql(database, table, df, auto_str=False, if_exists='fail'):
    '''Write the {df} to the {table} in the {database}.
    Specify if_exists='replace' to replace an existing table. Set auto_str to true
    to automatically turn non-numerical data to string.'''
    import sqlite3
    if auto_str:
        for li in df:
            dtype = str(df[li].dtype)
            if ('int' not in dtype) and ('float' not in dtype):
                df[li] = df[li].astype(str)
    conn = sqlite3.connect(database)
    try:
        conn.text_factory = lambda x: x.decode('utf-8', 'ignore')
        df.to_sql(table,conn,if_exists='fail')
    finally:
        conn.close()
    
def collect(filelist, file_reader, ignore_error=False, verbose=True):
    '''collect pieces of dataframe with the same variables and merge them together'''
    def read_file(fpath):
        df = None
        try:
            df = file_reader(fpath)
        except:
            if ignore_error==True:
                pass
        return df
    if verbose:
        print("Processing {} elements..".format(len(filelist)))
    r=[]
    len_last_sentence = 0
    for s,li in xenumerate(filelist):
        if verbose:
            print(' '*len_last_sentence, end='\r') # erase last printout
            s = "{} processing {}".format(s, li)
            len_last_sentence = len(s)
            print(s, end='\r')
        r.append(read_file(li))
    return r



from functools import wraps
from .pandas_internals import (register_series_accessor,
                              register_dataframe_accessor)


def register_dataframe_method(method):
    """Register a function as a method attached to the Pandas DataFrame.

    Example
    -------

    .. code-block:: python

        @register_dataframe_method
        def print_column(df, col):
            '''Print the dataframe column given'''
            print(df[col])
    """
    def inner(*args, **kwargs):

        class AccessorMethod(object):


            def __init__(self, pandas_obj):
                self._obj = pandas_obj

            @wraps(method)
            def __call__(self, *args, **kwargs):
                return method(self._obj, *args, **kwargs)

        register_dataframe_accessor(method.__name__)(AccessorMethod)

        return method

    return inner()


def register_series_method(method):
    """Register a function as a method attached to the Pandas Series.
    """
    def inner(*args, **kwargs):

        class AccessorMethod(object):
            __doc__ = method.__doc__

            def __init__(self, pandas_obj):
                self._obj = pandas_obj

            @wraps(method)
            def __call__(self, *args, **kwargs):
                return method(self._obj, *args, **kwargs)

        register_series_accessor(method.__name__)(AccessorMethod)

        return method

    return inner()



from .pandas_flavor import register_dataframe_accessor, register_series_accessor
from functools import wraps
from collections import OrderedDict
import pandas as pd

@register_dataframe_accessor('bs')
class DataFrameBooster:
    def __init__(self, obj):
        self._obj = obj

def register_dataframe_booster(name):
    def decorator(f):
        # hack
        @wraps(f)
        def wrapper(*args, **kwargs):
            instance = args[0]
            args = args[1:]
            return f(instance._obj, *args, **kwargs)
        setattr(DataFrameBooster, name, wrapper)
        return f
    return decorator

@register_series_accessor('bs')
class SeriesBooster:
    def __init__(self, obj):
        self._obj = obj

def register_series_booster(name):
    def decorator(f):
        # hack
        @wraps(f)
        def wrapper(*args, **kwargs):
            instance = args[0]
            args = args[1:]
            return f(instance._obj, *args, **kwargs)
        setattr(SeriesBooster, name, wrapper)
        return f
    return decorator

@register_dataframe_accessor('fm')
class FilterManager:
    """Manage boolean-array filters.
    
    """
    def __init__(self, obj):
        self._obj = obj
        self.__filters = OrderedDict()

    def add(self, msk, name):
        self.__filters[name] = msk

    def show(self):
        return self.__filters.keys()

    def combine(self):
        msk = pd.concat(self.__filters.values(), axis=1)
        return msk.all(axis=1)

    def clear(self):
        self.__filters = OrderedDict()

    def waterfall(self):
        """Show how base is reduced through each filter."""
        r = []
        totobs = self._obj.shape[0]
        r.append({'name': 'Base',
                  'nobs': totobs,
                  '%R': '',
                  '%T': ''})
        lastobs = totobs
        lastmsk = pd.Series([True]*self._obj.shape[0], index=self._obj.index)
        for name, msk in self.__filters.items():
            cursel = lastmsk & msk
            nobs = cursel.sum()
            r.append({'name': name,
                      'nobs': nobs,
                      '%R': nobs/lastobs,
                      '%T': nobs/totobs})
            lastobs = nobs
            lastmsk = cursel
        df = pd.DataFrame(r)
        df = df[['name', 'nobs', '%R', '%T']]
        return df

    def reduce(self):
        """Return records that pass all filters"""
        return self._obj.loc[self.combine()]

    def reduce_index(self):
        return self.reduce().index
